bilin.loop<-function(thecases=1:100,useTrueCov=FALSE,theseed=0,save=FALSE) {
  #Runs the bilinear algorithm for the plots in the paper.


  cat('--------------------------------------------------------\n')

  # Randomly/manually set seed for reproducibility
  # (omit 'theseed' for random seed; specify 'theseed' to start
  # from a specific seed)
  
  if (theseed != -1) {
    set.seed( theseed )
    cat('Using seed: ',theseed,'\n')
  }
  else {
    theseed <- floor(runif(1,0,1)*100000)
    cat('Using seed: ',theseed,'\n')
    set.seed( theseed )
  }

  #------------------------------------------------------------------------
  # Task 1: 
  # - Create model and data
  # - Learn models from eq class from 10 initial starting conditions
  # - Estimate which direct effects are determined, which are not
  # - Estimate value of determined direct effects (incl. zero / non-zero)
  # - Show statistics
  #------------------------------------------------------------------------

    # Problem parameters
    globalSanity <<- FALSE      # set to FALSE to do actual tests
    ncases <- length(thecases)               # number of cases to run (e.g. 100)

    # Solution parameters
    globalMaxIters <<- 20      # max iterations trying to find one solution
    globalTotTries <<- 10      # number of total tries
    globalSelTries <<- 5       # how many best tries to use
    globalDetStdev <<- 0.2
    globalZeroFact <<- 2

    # Show all settings:
    effSamples <- globalSamples
    if (useTrueCov) effSamples <- Inf
    cat('Settings:\n')
    cat('- useTrueCov:     ',useTrueCov,'\n',sep='')
    cat('- globalSamples:  ',effSamples,'\n',sep='')
    cat('- globalVars:     ',globalVars,'\n',sep='')
    cat('- globalExp:      ',globalExp,'\n',sep='')
    cat('- globalMaxIters: ',globalMaxIters,'\n',sep='')
    cat('- globalTotTries: ',globalTotTries,'\n',sep='')
    cat('- globalSelTries: ',globalSelTries,'\n',sep='')
    cat('- globalDetStdev: ',globalDetStdev,'\n',sep='')
    cat('- globalZeroFact: ',globalZeroFact,'\n',sep='')
    
    # This list will hold the determined status of the elements:
    # 1 for a direct effect determined to be non-zero, 0 for a 
    # determined zero, and -1 for underdetermined coefficients.
    Bdet <- list()

    # Similarly as above, but using only the Bij=0 constraints
    Bdetf <- list()
    
    # This list will hold the inferred models. Each element is the best
    # numerical estimate of the value of the corresponding coefficient.
    # Note that for those elements that are underdetermined (see above)
    # the value may not be very useful and may be completely off!
    # For those elements determined to be zero (see above), the value
    # here is zero.
    Bval <- list()

    # Loop over a number of cases to test
    casei <- 0
    cat('Working on the ',ncases,' cases:\n',sep='')
    for (i in thecases) {
    
      # Show progress (and wait a while so we can see it!)
      casei <- casei+1
      cat('[',casei,']',sep='')
    
      # Get the model and data, which is a list containing:
      # model: the true model (with B and Ce as list elements)
      # E: definition of experiments, each row is one experiment
      # Cx: list of true covariance matrices
      # X: list of sample matrices
      # Cx_sample: list of estimated covariance matrices
      D <- data.get(i)
            
      # Get the number of variables
      n<-ncol(D$E)

      if (useTrueCov)
      {
        # Do the infinite sample inference here
        resF <- bilin.new2.infinite.inference(D$E, D$Cx, D$model$B, onlyF=TRUE)
        res <- bilin.new2.infinite.inference(D$E, D$Cx, D$model$B)
      }
      else
      {
        # Do the finite sample inference here
        resF <- bilin.new2.finite.inference(D$E,D$Cx_sample, globalSamples, onlyF=TRUE)
        res <- bilin.new2.finite.inference(D$E, D$Cx_sample, globalSamples)            
      }
      
      # Store the result
      Bdet[[casei]] <- res$Bdet
      Bval[[casei]] <- res$Bval
      Bdetf[[casei]] <- resF$Bdet
      
    }

    # If useTrueCov then set globalSamples to Inf in display
    if (useTrueCov) globalSamples <<- Inf

    cat('Done!\n')
    
    # Print statistics of how well we did
    cat('Using only the zeros from indeptests:\n')
    data.stats(Bdetf,thecases)
    
    cat('Using the bilinear approach:\n')
    results<-data.stats(Bdet,thecases)
    #saving the results for plotting
    if ( save ) {
      if ( !useTrueCov ) {
	file=paste('results/bilin_',globalSamples,'.Rdata',sep='')
      } else {
	file=paste('results/bilin_',Inf,'.Rdata',sep='')
      }
      save(results,file=file)
    }
  cat('Used the seed: ',theseed,'\n')

}