
bilin.model.optimize <- function(Tall, tvecall, Ball, bvecall, Hbilinall)
{
  # Inputs:
  # Tall, tvecall, Ball, bvecall, Hbilinall
  #
  # These are linear constraints on the total and direct effects:
  # Tall %*% bilin.vectorize(Ttrue) == tvecall
  # Ball %*% bilin.vectorize(Btrue) == bvecall
  # Ttrue %*% (diag(n)-Btrue) == diag(n)
  # 
  # Does:
  # Starts from some random values for the total effects and the
  # direct effects, and minimizes the sum of squared errors in the
  # above equations by a simple alternating variables method.
  # Runs 'globalMaxIters' times and exists.
  #
  # Returns
  # A list res with:
  # res$B    - the B matrix at the end of the optimization
  # res$obj  - the objective function at the end
  #

  # Verbose (print out stuff) or not
  verb <- 0
  
  # Get the number of variables
  n <- sqrt(ncol(Tall))
  
  # Create initial values
  Tcur <- matrix(rnorm(n*n),n,n)
  Bcur <- matrix(rnorm(n*n),n,n)

  # Also get vectorized versions
  tcur <- bilin.vectorize(Tcur)
  bcur <- bilin.vectorize(Bcur)

  # Computing the objective the easy way
  obj1 <- sum(((Tcur %*% (diag(n) - Bcur)) - diag(n))^2)
  obj2 <- sum(((Tall %*% tcur) - tvecall)^2)
  obj3 <- sum(((Ball %*% bcur) - bvecall)^2)
  obj4 <- 0
  for (br in 1:(n*n))
  {
    for (tc in 1:(n*n))
    {
      if (Hbilinall[br,tc])
      {
        obj4 <- obj4 + bcur[br]*tcur[tc]
      }
    }
  }
  totobj <- obj1 + obj2 + obj3 + obj4
  if (verb) cat('Initial objective: ',totobj,'\n',sep='')
            
  # Alternating optimization of T and B
  for (iter in 1:globalMaxIters)
  {
    # Show progress
    if (verb) cat('[',iter,'] ',sep='')
    
    # --- Hold T fixed and optimize B ---
      
    H <- matrix(0,0,n*n)
    hvec <- matrix(0,0,1)
    
    # Matrix inverse constraint...
    for (i in 1:n)
    {
      for (j in 1:n)
      {
        targetval <- Tcur[i,j] - 1*(i==j)
        coeffs <- Tcur[i,]
        Bdummy <- matrix(0,n,n)
        Bdummy[,j] <- 1
        bdummy <- bilin.vectorize(Bdummy)
        inds <- which(bdummy==1)
        coeffsvec <- rep(0,n*n)
        coeffsvec[inds] <- coeffs
        H <- rbind(H, matrix(coeffsvec,1,n*n))
        hvec <- rbind(hvec, targetval)
      }
    }

    # Add the linear constraints on the direct effects
    H <- rbind(H, Ball)
    hvec <- rbind(hvec, matrix(bvecall,nrow(Ball),1))

    # Add the bilinear constraints
    for (br in 1:(n*n))
    {
      for (tc in 1:(n*n))
      {
        if (Hbilinall[br,tc])
        {
          newbc <- rep(0,n*n)
          newbc[br] <- tcur[tc]
          H <- rbind(H,newbc)
          hvec <- rbind(hvec,0)
        }
      }
    }
        
    # Solve using generalized inverse
    library("MASS")
    Hpinv <- bilin.ginvwrapper(H)
    bnew <- (Hpinv %*% hvec)
    bcur <- bnew
    Bcur <- bilin.matricize(bcur)
    
    # --- Hold B fixed and optimize T ---
    
    L <- matrix(0,0,n*n)
    lvec <- matrix(0,0,1)
      
    # Matrix inverse constraint...
    for (i in 1:n)
    {
      for (j in 1:n)
      {
        targetval <- 1*(i==j)
        Bdummy <- diag(n) - Bcur      
        coeffs <- Bdummy[,j]
        Tdummy <- matrix(0,n,n)
        Tdummy[i,] <- 1
        tdummy <- bilin.vectorize(Tdummy)
        inds <- which(tdummy==1)
        coeffsvec <- rep(0,n*n)
        coeffsvec[inds] <- coeffs
        L <- rbind(L, matrix(coeffsvec,1,n*n))
        lvec <- rbind(lvec, targetval)
      }
    }

    # Add the linear constraints on total effects
    L <- rbind(L, Tall)
    lvec <- rbind(lvec, matrix(tvecall,nrow(Tall),1))

    # Add the bilinear constraints
    for (br in 1:(n*n))
    {
      for (tc in 1:(n*n))
      {
        if (Hbilinall[br,tc])
        {
          newtc <- rep(0,n*n)
          newtc[tc] <- bcur[br]
          L <- rbind(L,newtc)
          lvec <- rbind(lvec,0)
        }
      }
    }
    
    # Solve using generalized inverse
    library("MASS")
    Lpinv <- bilin.ginvwrapper(L)
    tnew <- Lpinv %*% lvec
    tcur <- tnew
    Tcur <- bilin.matricize(tcur)
      
    # Compute and show objective function
    obj1 <- sum(((Tcur %*% (diag(n) - Bcur)) - diag(n))^2)
    obj2 <- sum(((Tall %*% tcur) - tvecall)^2)
    obj3 <- sum(((Ball %*% bcur) - bvecall)^2)
    obj4 <- 0
    for (br in 1:(n*n))
    {
      for (tc in 1:(n*n))
      {
        if (Hbilinall[br,tc])
        {
          obj4 <- obj4 + bcur[br]*tcur[tc]
        }
      }
    }
    totobj <- obj1 + obj2 + obj3 + obj4
    if (verb) cat(totobj,'\n',sep='')     
  }
  
  # Renormalize B to zero diagonal
  Bfinal <- diag(n) - bilin.normalizerow(diag(n) - Bcur)
  
  # Return the result
  res <- list()
  res$B <- Bfinal
  res$obj <- totobj
  res
    
}
