
bilin.new2.infinite.inference <- function(E, Cx, Btrue, onlyF=FALSE)
{
  # Inputs:
  #
  # E - a matrix with nexp rows and n columns, where nexp is the number
  #     of experiments and n is the number of variables. an entry is 
  #     equal to 1 if intervened, 0 if passively observed, and -1 if 
  #     latent.
  #
  # Cx - a list of covariance matrices. Cx[[1]] contains the covariance
  #      matrix for the first experiment, Cx[[2]] for the second 
  #      experiment, and so on, with NA for all elements where
  #      at least one of the variables is not observed.
  #
  # Btrue - True matrix B. This is only used to verify
  #         the constraints provided by independence test code
  #
  # onlyF - if TRUE, it only gives the results obtained with
  #         independence-test code
  #
  # Outputs:
  # 
  # res$Bdet - an n-by-n connection matrix, where B[i,j] is the connection 
  #            from node j to node i, and the elements are 1 if inferred to
  #            be definitely non-zero, 0 if inferred to be definitely zero,
  #            and -1 if underdetermined by this procedure.
  #
  # res$Bval - The n-by-n connection matrix with best estimates of the
  #            coefficients.
  #

  # Verbose?
  verb <- 0
  
  #-------------------------------------------------------------
  # Sourcing independence test code (and anything else needed)
  #-------------------------------------------------------------

  # Independence test code
#  bilin.sourceDir("./../../indepedence/", trace=FALSE)
  
  # Other functions needed here
#  source('../isIndependent.R')
#  source('../../experiments/pair2lin.R')
  
  #-------------------------------------------------------------
  # Some simple initializations
  #-------------------------------------------------------------

  # Number of variables
  n <- ncol(E)
  
  # Change notation of the E matrix:
  E <- t(E+1)
  nexp <- ncol(E)

  #-------------------------------------------------------------
  # Run independence test code for finding faithfulness constraints!
  #-------------------------------------------------------------

  # This will store all the constraints on the total effects
  Tall <- matrix(0,0,n*n)
  tvecall <- rep(0,0)
  
  # This will store all the constraints on the direct effects
  Ball <- matrix(0,0,n*n)
  bvecall <- rep(0,0)

  # This will store the bilinear constraints
  Hbilinall <- matrix(FALSE,n*n,n*n)

  # We will do this separately for each experiment
  for (ei in 1:nexp)
  {
    # Call independence test code
    exp <- E[,ei]
    Call <- Cx[[ei]]
    nsamples <- Inf
    maxCsets <- 10000   # limiting size of conditioning set? no.
    signif1 <- 0.1    # no use, since using Inf sample size
    signif2 <- signif1

    res = faithConst4Bilin(exp, Call, nsamples, maxCsets, signif1, signif2)

    # Get the output
    T <- res[[3]]
    tvec <- res[[4]]
    B <- res[[1]]
    bvec <- res[[2]]
    Hbilin <- res[[5]]
    
    # Verify the constraints provided by independence test procedure.
    verify <- TRUE
    if (verify)
    {  
      truevals <- B %*% bilin.vectorize(Btrue)
      if (length(bvec)>0)
      {  
        if (max(abs(truevals-bvec))>1e-8) cat('PROBLEM ON B!')
      }
      Ttrue <- solve(diag(n)-Btrue)
      truevals <- T %*% bilin.vectorize(Ttrue)
      if (length(tvec)>0)
      {  
        if (max(abs(truevals-tvec))>1e-8) cat('PROBLEM ON T!')
      }
      
      for (tr in 1:nrow(Hbilin))
      {
        for (tc in 1:ncol(Hbilin))
        {
          if (Hbilin[tr,tc])
          {
            Ttruevec <- bilin.vectorize(Ttrue)
            Btruevec <- bilin.vectorize(Btrue)
            if (abs(Ttruevec[tc] * Btruevec[tr])>1e-8)  # correct
            {
              stop('Found an error!')
            }
          }
        }
      }
    }
    
    # Store/add the results to what we have previously
    Tall <- rbind(Tall,T)
    tvecall <- c(tvecall,tvec)
    Ball <- rbind(Ball,B)
    bvecall <- c(bvecall,bvec)
    Hbilinall <- Hbilin | Hbilinall
    
  }

  #-------------------------------------------------------------
  # if onlyF is TRUE, we only get the zeros from independence tests.
  # Construct the thing to return here, and leave.
  #-------------------------------------------------------------
  if (onlyF)
  {
    if (nrow(Ball)==0)
    {
      Bdet <- matrix(-1,n,n)
      diag(Bdet) <- 0
      res <- list()
      res$Bdet <- Bdet
      return(res)
    }
      
    Bknownzeros <- rep(0,n*n)
    for (i in 1:nrow(Ball))
    {
      for (j in 1:(n*n))
      {    
        testvec <- rep(0,n*n)
        testvec[j] <- 1
      
        if (max(abs(testvec - Ball[i,]))<1e-8)
        {
          if (abs(bvecall[i]) < 1e-8)
          {
            Bknownzeros[j] <- 1
          }
        }
      }
    }
    Bknownzeros <- bilin.matricize(Bknownzeros)
    diag(Bknownzeros) <- 1
    Bdet <- (1-Bknownzeros)*(-1)

    # Finally, return Bdet and Bval. 
    res <- list()
    res$Bdet <- Bdet
    return(res)
    
  }
  
  #-------------------------------------------------------------
  # Reduce the covariance matrices to only include the 
  # intervened and observed variables, i.e. not the hidden
  # ones (that only contain NA anyway)
  #-------------------------------------------------------------

  # Go through all experiments  
  for (i in 1:nexp)
  {
    # Which are the intervened and observed vars?
    thesevars <- which(E[,i] > 0)
  
    # Get the relevant data
    Cx[[i]] <- Cx[[i]][thesevars,thesevars]    
  }

  #-------------------------------------------------------------
  # Create the constraints on the total effects. 
  #-------------------------------------------------------------

  # Create the constraints
  Kfull <- bilin.createConstraints(E, Cx)

  # Verify them!
  verify <- TRUE
  if (verify)
  {
    if (max(abs(Kfull %*% bilin.vectorize(Ttrue)))>1e-8)
    {
      cat('Error in Kfull constraints!')
    }
  }

  # Add in the Kfull constraints to Tall
  Tall <- rbind(Tall,Kfull)
  tvecall <- c(tvecall,rep(0,nrow(Kfull)))
  
  #-------------------------------------------------------------
  # At this point, we have 4 sources of information on the
  # true model. We have:
  # Tall %*% bilin.vectorize(Ttrue) == tvecall
  # Ball %*% bilin.vectorize(Btrue) == bvecall
  # Ttrue %*% (diag(n)-Btrue) == diag(n)
  # if (Hbilinall[i,j])
  #   bilin.vectorize(Btrue)[i] * bilin.vectorize(Ttrue)[j] == 0  
  #-------------------------------------------------------------
  
  # This will store the objective functions
  objs <- rep(0,globalTotTries)

  # This will store the corresponding models
  Bmat <- list()

  # Do a total of 'globalTotTries' optimizations
  for (tryiter in 1:globalTotTries)
  {
    # Show progress
    if (verb) cat('- Optimizing ',tryiter,' of ',globalTotTries,': ',sep='')
    
    # Run the optimization
    res <- bilin.model.optimize(Tall, tvecall, Ball, bvecall, Hbilinall)

    # Store the results
    objs[tryiter] <- res$obj
    Bmat[[tryiter]] <- res$B    
  }

  # Select the 'globalSelTries' best results
  res <- sort.int(objs, index.return=TRUE)
  bestid <- res$ix[1:globalSelTries]
  Bsel <- list()
  objsel <- rep(0,globalSelTries)
  for (i in 1:globalSelTries)
  {
    Bsel[[i]] <- Bmat[[bestid[i]]]
    objsel[i] <- objs[bestid[i]]
  }
  
  # Matrix with all the selected estimates
  Bselvec <- matrix(0,0,n*n)
  for (i in 1:globalSelTries)
  {
    Bselvec <- rbind(Bselvec,matrix(bilin.vectorize(Bsel[[i]]),1,n*n))
  }

  # Compute means and standard deviations
  Bselvecmean <- colMeans(Bselvec)
  Bselvecstdev <- sqrt(diag(var(Bselvec)))
  
  # Infer which are determined by the global threshold
  Bdetvec <- (Bselvecstdev < globalDetStdev)

  # Infer which of the determined ones are zero
  Biszerovec <- rep(0,n*n)
  for (i in 1:(n*n))
  {
    if (Bdetvec[i] && (abs(Bselvecmean[i]) <= globalZeroFact*Bselvecstdev[i]))
    {
      Biszerovec[i] <- 1
    }
  }
  
  # Construct the final output: Bdet and Bval. Bdet has elements:
  # 1: determined non-zero, 0: determined zero, -1: underdetermined,
  # with a diagonal fixed at 0. Bval has the best guess for all
  # elements (non-zero values for those determined as non-zero,
  # 0 for those determined as zero, and arbitrary values for those
  # underdetermined. The diagonal is again 0.
  Bvalvec <- rep(0,n*n)
  for (i in 1:(n*n))
  {
    if (Bdetvec[i] && Biszerovec[i])
    {
      Bdetvec[i] <- 0
      Bvalvec[i] <- 0
    }
    else if (Bdetvec[i])
    {
      Bdetvec[i] <- 1
      Bvalvec[i] <- Bselvecmean[i]
    }
    else
    {
      Bdetvec[i] <- -1
      Bvalvec[i] <- Bselvecmean[i]
    }
  }
  Bdet <- bilin.matricize(Bdetvec)
  Bval <- bilin.matricize(Bvalvec)
  
  # Finally, return Bdet and Bval. 
  res <- list()
  res$Bdet <- Bdet
  res$Bval <- Bval
  res
  
}
