

# This algorithm returns for each pair of variables x, y all minimal Sepsets C such that 
# x is independent of y given C. The algorithm iterates through potential Sepsets in order of size,
# starting with the empty set. It is similar to the SGS algorithm, except that when it finds 
# a Sepset of size k, it continues checking all Sepsets of size k before moving to the next pair of 
# variables.

# INPUT:
# Call: covariance matrix over ALL n variables
# nhidi: vector containing the indices of the variables that are not hidden
# nsamples: number of samples
# maxCsets: maximum size of the conditioning set for minimal independence tests
# significance: level of significance to establish independence

# OUTPUT:
# returns a n x n list "sepsets" with
#     sepsets[[i]][[j]] = sepsets[[j]][[i]] and
#     sepsets[[i]][[i]] = NULL for all i
#     for x_i, x_j, sepsets[[i]][[j]] is a list of sets of variables C such that x_||_y | C.
#          If C = emptyset, then sepsets[[i]][[j]] = 0
#          If no such C was found, then sepsets[[i]][[j]]=NULL.



allminSepsets = function(Call, nhidi, nsamples, maxCsets, significance) {

  n = ncol(Call)	 
  nnothid = length(nhidi)   
  
                                        # initialize an n x n matrix as a skeleton, diagonal zero, otherwise 1
  
  skel = matrix(1, nrow=n, ncol=n)
  diag(skel) <- 0  
  
                                        # initialize the list of sepsets as an n x n list, which contains for each
                                        # pair of variables an initially empty list of the minimal separating sets
  sepset = vector(length = n, mode = "list")
  for (i in 1:n) {
    sepset[[i]] = vector(length = n, mode = "list")
    for(j in 1:n){sepset[[i]][[j]] = list()}
  }
  
  if(nnothid<2){sepset}else{

                                        # pass through all symmetric pairs of variables, test if they are
                                        # unconditionally independent

    for (i in 1:(nnothid-1)) {
      for (j in (i+1):nnothid) {
    	x = nhidi[i]
	y = nhidi[j]
	
	pval = isIndependent(Call, x, y, given = NULL, nsamples)
      	if (pval>significance) { 
          skel[x,y] = 0
          skel[y,x] = 0 
          sepset[[x]][[y]][[1]] = 0 # mark the separating set as empty set by 0
          sepset[[y]][[x]][[1]] = 0 
        }
      }
    }

                                         # pass through all symmetric pairs of variables and test for conditional independence

    for (i in 1:(nnothid-1)) {
      for (j in (i+1):nnothid) {
    	x = nhidi[i]
	y = nhidi[j]
	css=1	# size of conditioning set
    	while(skel[x,y]!=0 && css<=(nnothid-2) && css<maxCsets){	   # if the variables have not yet been found to be independent
          pcsv = setdiff(nhidi, c(x,y))    	   #potential conditioning set vars

          pss = if (length(pcsv) == css) matrix(pcsv) else combn(pcsv, css) # potential sepsets of size css
          for(idx in 1:ncol(pss)){       		   # consider all possible sepsets
            pval = isIndependent(Call, x, y, given = pss[,idx], nsamples)
            if (pval>significance) {
              skel[x,y] = 0
              skel[y,x] = 0
              sepset[[x]][[y]][[length(sepset[[x]][[y]])+1]] = pss[,idx] 
              sepset[[y]][[x]][[length(sepset[[y]][[x]])+1]] = pss[,idx]
            }
          }
          css=css+1
        }
      }
    }
    sepset
  }
}


