
# Given the covariance matrix this method generates constraints that are implied by the 
# assumption of faithfulness in this experiment:
# - linear constraints on the total effects,
# - linear constraints on the direct effects 
# - bilinear constraints that are a product of a total and a direct effect
# These constraints are used in the BILIN algorithm


# INPUT
# exp: vector of length n consisting of 0,1,2 representing an experiment (hidden, obs, interv)
# Call: n x n covariance matrix for ALL variables in V for the exp-eriment
# nsamples: number of samples
# maxCsets: maximum size of the conditioning set for minimal independence tests (the procedure 
# 	    	    will still check for dependence using conditioning sets of size maxCsets+1 when possible)
# signif1: level of significance to establish independence: pval>signif1
# signif2: level of significance to establish dependence: pval<=signif2
# verbose: boolean controlling the commentary


# OUTPUT (see Eq 4 in paper)
# matrix K1 with n^2 columns of linear constraints on direct effects
# vector k1 corresponding to the direct effects equations (currently only zeros)
# matrix K2 with n^2 columns of linear constraints on total effects
# vector k2 corresponding to the total effects equations (currently only zeros) 
# logical matrix H of size n^2 x n^2 with entries that are TRUE iff the product of a direct and a total 
# effect is known to be 0


faithConst4Bilin = function(exp, Call, nsamples, maxCsets, signif1, signif2, verbose = FALSE){

  n = length(exp)
  evec <- (exp==2)
  eveci <- which(evec)
  evecobs <- (exp==1)
  evecobsi <- which(evecobs)
  evecnothidden <- (exp>0)
  evecnothiddeni <- which(evecnothidden)    
  evechidden <- (exp == 0) 
  evechiddeni <- which(evechidden) 
  
                                        # Generate the index matrix for use in accessing columns of the constraint matrices K1, K2 and H
  
  P <- matrix(0,2,0)
  for (i in 1:n){
    for (j in 1:n){
      P <- cbind(P,c(i,j))
    }
  }
  
                                        # Initialize constraint matrices for the output (rows will be added with rbind)
  K1 <- matrix(0,0,n*n) # constraints on direct effects
  K2 <- matrix(0,0,n*n) # constraints on total effects
  H <- matrix(FALSE,n^2,n^2) #bilinear constraints
  
                                          # get all the conditioning sets that make a pair of variables (conditionally) independent
  sepsets = allminSepsets(Call, evecnothiddeni, nsamples, maxCsets, signif1)
  
                                        # consider independencies x_||_y | C with x,y in U first
  if(length(evecobsi)>=2){
    
    for(i in 1:(length(evecobsi)-1)){
      for(j in (i+1):length(evecobsi)){
        x = evecobsi[i]
        y = evecobsi[j]
        
                                        # consider all possible sepsets of x and y, if any	    
        if(length(sepsets[[x]][[y]])>0){
          
          for(ssi in 1:length(sepsets[[x]][[y]])){
            
            if(length(sepsets[[x]][[y]][[ssi]])==1 && (sepsets[[x]][[y]][[ssi]]) == 0){
              
                                        # x and y are unconditionally independent
              K1 = addSimpleConst(x,y,K1,P,msg= 'b(x-->y)=0', verbose = verbose)
              K1 = addSimpleConst(y,x,K1,P,msg='b(y-->x)=0', verbose = verbose)
              K2 = addLongConst(x, y, eveci, K2, Call, P,msg='t(x=>y||J cup x)=0', verbose = verbose)
              K2 = addLongConst(y, x, eveci, K2, Call, P,msg='t(y=>x||J cup y)=0', verbose = verbose)
              
              if(length(eveci)==0){						# pass obs data set
                for(u in c(evechiddeni, setdiff(evecobsi, c(x,y)))){
                  H = addBilin(H,x,u,u,y,P, msg='b(x->u) t(u=>y)=0', verbose = verbose)
                  H = addBilin(H,u,y,x,u,P, msg='b(u->y) t(x=>u)=0', verbose = verbose)
                  H = addBilin(H,y,u,u,x,P, msg='b(y->u) t(u=>x)=0', verbose = verbose)
                  H = addBilin(H,u,x,y,u,P, msg='b(u->x) t(y=>u)=0', verbose = verbose)
                  H = addBilin(H,u,y,u,x,P, msg='b(u->y) t(u=>x)=0', verbose = verbose)
                  H = addBilin(H,u,x,u,y,P, msg='b(u->x) t(u=>y)=0', verbose = verbose)
                }
              }else{
                for(xj in eveci){
                  K1 = addSimpleConst(xj,x,K1,P,val=Call[y,xj],msg='b(xj->x)t(xj=>y||J)=0', verbose = verbose)
                  K1 = addSimpleConst(xj,y,K1,P,val=Call[x,xj],msg='b(xj->y)t(xj=>x||J)=0', verbose = verbose)
                }
              }
              
                                        # check for colliders w in U between x and y
              for(w in setdiff(evecobsi, c(x,y))){
                
                pval = isIndependent(Call, x, y, given = w, nsamples)
                if(pval<=signif2){
					# x and y are unconditionally independent, but dependent conditional 
					# on some w in U
                  K1 = addSimpleConst(w,x,K1,P,msg='b(w->x)=0', verbose = verbose)
                  K1 = addSimpleConst(w,y,K1,P,msg='b(w->y)=0', verbose = verbose)
                  K2 = addLongConst(w,x,eveci,K2,Call,P,msg='t(w=>x||J cup w)=0', verbose = verbose)
                  K2 = addLongConst(w,y,eveci,K2,Call,P,msg='t(w=>y||J cup w)=0', verbose = verbose)
                  
                  if(length(eveci)==0){							# pass obs data set
                    for(u in c(evechiddeni, setdiff(evecobsi, c(x,y,w)))){
                      H = addBilin(H,w,u,u,x,P, msg='b(w->u) t(u=>x)=0', verbose = verbose)
                      H = addBilin(H,w,u,u,y,P, msg='b(w->u) t(u=>y)=0', verbose = verbose)
                      H = addBilin(H,u,x,w,u,P, msg='b(u->x) t(w=>u)=0', verbose = verbose)
                      H = addBilin(H,u,y,w,u,P, msg='b(u->y) t(w=>u)=0', verbose = verbose)
                    }
                  }
                }
                
                if(pval>signif1 && length(eveci)==0){
                  
					#  x and y are unconditionally independent, and remain independent conditional on some w in U
                  H = addBilin(H,x,w,y,w,P, msg='b(x->w) t(y=>w)=0', verbose = verbose)
                  H = addBilin(H,y,w,x,w,P, msg='b(y->w) t(x=>w)=0', verbose = verbose)
                }
              }
            }else{		        # x and y are conditionally independent
              
              cset = (sepsets[[x]][[y]][[ssi]])
              
              K1 = addSimpleConst(x,y,K1,P,msg='C: b(x->y)=0', verbose = verbose)
              K1 = addSimpleConst(y,x,K1,P,msg='C: b(y->x)=0', verbose = verbose)
              
              if(all(cset %in% eveci)){				#cset in J	
                K2 = addLongConst(x, y, eveci, K2,Call, P,msg='C: t(x=>y||J cup x)=0', verbose = verbose)
                K2 = addLongConst(y, x, eveci, K2,Call, P,msg='C: t(y=>x||J cup y)=0', verbose = verbose)
              }
              
                                        # check for colliders w in U\C between x and y
              for(w in setdiff(evecobsi, c(cset,x,y))){
                pval = isIndependent(Call, x, y, given = c(cset,w), nsamples)		
                if(pval<=signif2){
                  
                                        # x and y are independent conditional on cset, but dependent conditional on cset \cup w
                  K1 = addSimpleConst(w,x,K1,P,msg='C: b(w->x)=0', verbose = verbose)
                  K1 = addSimpleConst(w,y,K1,P,msg='C: b(w->y)=0', verbose = verbose)
                  K2 = addLongConst(w, x, eveci, K2, Call, P,msg='C: t(w=>x||J cup w)=0', verbose = verbose)
                  K2 = addLongConst(w, y, eveci, K2, Call, P,msg='C: t(w=>y||J cup w)=0', verbose = verbose)
                  
                  for(c in setdiff(cset,eveci)){
                    K1 = addSimpleConst(w,c,K1,P,msg='C: b(w=>c)=0', verbose = verbose)
                    K2 = addLongConst(w, c, eveci, K2, Call, P,msg='C: t(w=>c||J cup w)=0', verbose = verbose)
                  }					
                  
                  if(length(eveci)==0){				#pass obs 
                    for(u in c(evechiddeni, setdiff(evecobsi, c(cset,x,y,w)))){
                      H = addBilin(H,w,u,u,x,P, msg='b(w->u) t(u=>x)=0', verbose = verbose)
                      H = addBilin(H,w,u,u,y,P, msg='b(w->u) t(u=>y)=0', verbose = verbose)
                      H = addBilin(H,u,x,w,u,P, msg='b(u->x) t(w=>u)=0', verbose = verbose)
                      H = addBilin(H,u,y,w,u,P, msg='b(u->y) t(w=>u)=0', verbose = verbose)
                      
                      for(c in setdiff(cset,eveci)){
                        H = addBilin(H,w,u,u,c,P, msg='b(w->u) t(u=>c)=0', verbose = verbose)
                        H = addBilin(H,u,c,w,u,P, msg='b(u->c) t(w=>u)=0', verbose = verbose)
                      }		    
                    }
                  }    		
                }
                
                                        #  x and y are conditionally independent, and remain independent conditional on cset cup w for some w in U
                if(pval>signif1){
                                        # bilinear constraint not representable
                }
              }
            }
          }
        }
      }
    }
  }
      
                                        # consider x_||_y | C when x in J and y in U
  if(length(eveci)>0 && length(evecobsi)>0){
    
    for(i in 1:length(eveci)){
      for(j in 1:length(evecobsi)){
        x = eveci[i]
        y = evecobsi[j]
        
                                        # consider all possible sepsets of x and y, if any	    
        if(length(sepsets[[x]][[y]])>0){
          
          for(ssi in 1:length(sepsets[[x]][[y]])){
            
            if( length(sepsets[[x]][[y]][[ssi]])==1 && (sepsets[[x]][[y]][[ssi]]) == 0){
              
                                        # x and y are unconditionally independent
              K1 = addSimpleConst(x,y,K1,P,msg='*b(x->y)=0', verbose = verbose)
              K2 = addLongConst(x, y, setdiff(eveci,x), K2, Call, P,msg='*t(x=>y||J)=0', verbose = verbose)	# standard experimental effect constraint
              
              if(length(eveci)==1){			# only x is intervened
                for(u in c(evechiddeni,setdiff(evecobsi, y))){
                  H = addBilin(H,u,y,x,u,P, msg='b(u->y) t(x=>u)=0', verbose = verbose)
                }
              }
              
                                        # check for colliders w in U between x and y
              for(w in setdiff(evecobsi, y)){
                pval = isIndependent(Call, x, y, given = w, nsamples) 
                if(pval<=signif2){
                  
					# x and y are unconditionally independent, but dependent conditional on w
                  K1 = addSimpleConst(w,y,K1,P,msg='*b(w->y)=0', verbose = verbose)
                  K2 = addLongConst(w, y, eveci, K2, Call, P,msg='*t(w=>y||J cup w)=0', verbose = verbose)
                  
    					# bilinear constraint not representable: t(w->u||J \cup w) t(u->y||J \cup u) = 0 
                }
                
                if(pval>signif1 && length(eveci)==1){	    # only x is intervened
                  
					#  x and y are unconditionally independent, and remain independent conditional on some w in U
                  H = addBilin(H,y,w,x,w,P, msg='b(y->w) t(x=>w)=0', verbose = verbose)
                }
              }
            }else{			# x and y are independent conditional on cset
              
              cset = (sepsets[[x]][[y]][[ssi]])
              
              K1 = addSimpleConst(x,y,K1,P,msg='C: *b(x->y)=0', verbose = verbose)		
              
                                        # bilinear constraints not representable: 
                                        # t(x->u||J \cup C \cup x) t(u->y||J \cup C \cup u) = 0
                                        # t(u-> y|| J cup C cup u ) t(u -> x || J cup C cup u ) = 0
              
              if(length(cset)==1){			#cset is a singleton conditioning set
                c = cset[1]
                if(exp[c]==1){ 
                  
                  K1 = addSimpleConst(y,c,K1,P,msg='C: *b(y->c)=0', verbose = verbose)
                  K2 = addLongConst(y,c,eveci,K2,Call, P,msg='C: *t(y=>c||J cup y)=0', verbose = verbose)
                  
                                        # bilinear constraint not representable: t( y -> u || J cup y) t( u -> c || J cup u)=0
                }
              }
              
                                        # check for colliders w in U between x and y
              for(w in setdiff(evecobsi, c(cset,y))){
                
                pval = isIndependent(Call, x, y, given=c(cset,w), nsamples)		
                if(pval<=signif2){
                  
                                        # x and y are independent conditional on cset, but dependent conditional on cset \cup w
                  K1 = addSimpleConst(w,y,K1,P,msg='C: *b(w->y)=0', verbose = verbose)
                  K2 = addLongConst(w,y,eveci,K2,Call, P,msg='C: *t(w=>y||J cup w)=0', verbose = verbose)
                  
                  for(c in setdiff(cset, eveci)){
                    K1 = addSimpleConst(w,c,K1,P,msg='C: *b(w->c)=0', verbose = verbose)
                    K2 = addLongConst(w,c,eveci,K2,Call, P,msg='C: *t(w=>c||J cup w)=0', verbose = verbose)
                  }
                  
                                        # bilinear constraints not representable: 
                                        # t( w -> u || J cup w ) t(u -> y || J cup u) = 0
                                        # for c in C: t( w -> u || J cup w ) t(u -> c || J cup u) = 0 		
                }
                
                
                
                if(pval>signif1){
                  
                                        #  x and y are conditionally independent, and remain independent conditional cset cup w for some w in U
                                        # bilinear constraint not representable: t(x=>w||J cup C cup x) t(y=>w||J cup C cup y) = 0
                  
                }
              }
            }
          }
        }
      }

        # for(j in setdiff((1:length(eveci)),i)){
	# 
	# 	# x and y are both intervened, and independent conditional on some w in U
	# 	# this step is redundant if experimental effects are picked up elsewhere
 	# 
	# 	x = eveci[i]
	# 	y = eveci[j]
	# 
	# 	for(w in evecobsi){
	# 	      	pval =   isIndependent(Call, x, y, given=w, nsamples)		
	# 		if(pval>signif1){
	# 		
	# 			# bilinear constraint not representable t(x=>w||J cup C cup x) t(y=>w||J cup C cup y) = 0
	# 		
	# 		}
	# 	}
	# }      
    }
  }
  list(K1, rep(0, nrow(K1)), K2, rep(0,nrow(K2)), H)
}









# adds a constraint d(from-->to)=0 to the constraint matrix
# (can be used for total and for direct effect constraints)

# INPUT
# from: origin variable in direct/total effect 
# to:  end variable in direct/total effect 
# constMat: matrix to which the constraint should be added (K1 or K2)
# P: matrix to look up indices
# val: covariance t(xj=>y||J) used for constraints of the form b(xj->x)t(xj=>y||J)=0
# msg: note for debugging
# verbose: boolean to regulate output of msg

# OUTPUT
# constMat: the original constraint matrix with the new constraint added in a new row

addSimpleConst = function(from, to, constMat, P, val = 1, msg=NA, verbose = FALSE){
  nn <- ncol(constMat)
  const <- rep(0, nn)
  const[bilin.pair2lin(P,from,to)] <- val #Antti added bilin. in front of pair2lin
  constMat <- rbind(constMat, const)
  
  if ( verbose && !is.na(msg) ) {
    cat('%%',msg,'\n')
  }
  
  constMat
}









# adds total effect constraints t(from=>to||J cup from)=0 to the constraint matrix K2
# using Eqn 2

# INPUT
# from: origin variable of the set of paths considered in the constraint
# to:  end variable of the set of paths considered in the constraint
# xjs: variables in the intervention set
# constMat: matrix to which the constraint should be added (K1 or K2)
# Call: covariance matrix of ALL variables in V in this experiment
# P: matrix to look up indices
# msg: note for debugging
# verbose: boolean to regulate output of msg

# OUTPUT
# constMat: the original constraint matrix with the new constraint added in a new row

addLongConst = function(from, to, xjs, constMat, Call, P, msg=NA, verbose = FALSE){

  const = rep(0,ncol(constMat))
  const[bilin.pair2lin(P,from,to)] = 1
  
  for(xj in xjs){
    const[bilin.pair2lin(P,from,xj)] = -Call[to,xj]
  }
  constMat = rbind(constMat, const)
  
  if ( verbose && !is.na(msg) ) {
    cat('%%',msg,'\n')
  }
  constMat
}




# sets entry in H to TRUE for bilinear constraint b(x->y) t(w=>v) = 0 

# INPUT
# H: boolean matrix indicating which products of direct x total effects are zero
# fromb: origin variable in direct effect
# tob: end variable in direct effect
# fromt: origin variable in total effect
# tot: end variable in total effect
# P: matrix to look up indices
# msg: note for debugging
# verbose: boolean to regulate output of msg

# OUTPUT
# H: boolean matrix with the new bilinear constraint added

addBilin = function(H,fromb,tob,fromt,tot,P, msg=NA, verbose = FALSE){

  H[bilin.pair2lin(P,fromb,tob), bilin.pair2lin(P,fromt,tot)] = TRUE
  
  if ( verbose && !is.na(msg) ) {
    cat('%%',msg,'\n')
  }
  H
}



