
# Uses the faithfulness rules described in Hyttinen et al (2010) to determine which experimental effects and which
# products of experimental effects are zero. The results are used in the HEH algorithm.
# Basically, this is a subset of the inferences done by faithConst4Lininf(...)


# INPUT
# Z: square matrix with n(n-1)2^(n-2) columns and rows 
#    Z[i,j] = 1 for those entries for which it is known that the product of the experimental effects 
#    represented by the row and column indices is zero, and which is zero otherwise
# z: ( n(n-1)2^(n-2))-ary vector which is 1 if the exp effect of the entry is known to be zero, and zero otherwise
# Call: covariance matrix over all variables in V for the exp-eriment
# exp: n-ary vector of {0,1,2} indicating the role of each variable in the experiment (hidden, obs, interv)
# nsamples: number of samples in exp
# maxCsets: maximum size of the conditioning set for minimal independence tests (the procedure 
# 	    	    will still check dependence using conditioning sets of size maxCsets+1 when possible)
# signif1: level of significance to establish independence: pval>signif1
# signif2: level of significance to establish dependence: pval<=signif2
# verbose: boolean controlling the commentary

# OUTPUT:
# returns a list containing Z and z after they have been updated using the implications from faithfulness 
# derived from this experiment, as specified in Hyttinen et al (2010)




faithConst4Heh = function(Z, z, Call, exp, nsamples, maxCsets, signif1, signif2,verbose=FALSE){
  n = length(exp)
  evec <- (exp==2)
  eveci <- which(evec)
  evecobs <- (exp==1)
  evecobsi <- which(evecobs)
  evecnothidden <- (exp>0)
  evecnothiddeni <- which(evecnothidden)  
  evechidden <- (exp == 0) 
  evechiddeni <- which(evechidden) 
  
  T = createParameterTable(n) # lookup table to find index of a particular experimental effect
  
                                        # get all the conditioning sets that make a pair of variables (conditionally) independent
  sepsets = allminSepsets(Call, evecnothiddeni, nsamples, maxCsets, signif1)
  
                                        # consider independencies x_||_y | C with x,y in U first
  if(length(evecobsi)>=2){
    
    for(i in 1:(length(evecobsi)-1)){
      for(j in (i+1):length(evecobsi)){
        x = evecobsi[i]
        y = evecobsi[j]
        
                                        # consider all possible sepsets of x and y, if any	    
        
        if(length(sepsets[[x]][[y]])>0){
          
          for(ssi in 1:length(sepsets[[x]][[y]])){
            
            
            if(length(sepsets[[x]][[y]][[ssi]])==1 && (sepsets[[x]][[y]][[ssi]]) == 0){
              
                                        # x and y are unconditionally independent
              
              z = add2zvec(z,x,y,setdiff(1:n, y),T,n, 'skel1', verbose) 
              z = add2zvec(z,y,x,setdiff(1:n, x),T,n, 'skel2', verbose)
              
              
                                        # the case when x and y are unconditionally independent, but dependent conditional on some w in U
                                        # is not considered in HEH

            }else{		# x and y are conditionally independent
    
              cset = (sepsets[[x]][[y]][[ssi]])
              
              z = add2zvec(z,x,y,setdiff(1:n,y),T,n, 'skel3', verbose)
              z = add2zvec(z,y,x,setdiff(1:n,x),T,n, 'skel4', verbose)
              
              
                                        # the case when x and y are independent conditional on cset, but dependent conditional on cset \cup w
                                        # is not considered in HEH
            }
          }
        }
      }
    }
  }

                                        # consider x_||_y | C when x in J and y in U
  if(length(eveci)>0 && length(evecobsi)>0){
    
    for(i in 1:length(eveci)){
      for(j in 1:length(evecobsi)){
        x = eveci[i]
        y = evecobsi[j]
    
                                        # consider all possible sepsets of x and y, if any	    
    
        if(length(sepsets[[x]][[y]])>0){
    
          for(ssi in 1:length(sepsets[[x]][[y]])){
    
            if( length(sepsets[[x]][[y]][[ssi]])==1 && (sepsets[[x]][[y]][[ssi]]) == 0){
    
                                        # x and y are unconditionally independent
                        
              z = add2zvec(z,x,y,setdiff(1:n,y),T,n, 'skel5', verbose)

              for(u in setdiff(evecobsi, y)){
                Z = add2Zmat(Z,x,u,eveci,u,y,setdiff(1:n, y),T,n, 'HEH Rule 1', verbose)
              }
    
                                        # the case when x and y are unconditionally independent, but dependent conditional on w
                                        # is not considered in HEH

                                        # instead the a specific case is considered in HEH:
                                        # x and y are unconditionally independent, and independent conditional on w: HEH Rule 2
              for(w in setdiff(evecobsi, y)){
                pval = isIndependent(Call, x, y, given = w, nsamples) 
                if(pval>signif1){
                  Z = add2Zmat(Z,x,w,eveci,y,w,setdiff(1:n, w),T,n, 'HEH Rule 2 sp', verbose)
					#  t(x=>w||J) b(y->w) = 0; no collider at w
                }
              }
			
    			
            }else{			# x and y are independent conditional on cset
    
              cset = (sepsets[[x]][[y]][[ssi]])
		
              z = add2zvec(z,x,y,setdiff(1:n, y),T,n, 'skel6', verbose)    

              if(length(cset)==1){			#cset is a singleton conditioning set
                c = cset[1]
                if(exp[c]==1){ 	                # c is observed
                  z = add2zvec(z,y,c,setdiff(1:n, c),T,n, 'HEH Rule 2', verbose)
                }
              }

    		# the case where x and y are independent conditional on cset, but dependent conditional on cset \cup w
		# is not considered in HEH
            }
          }
        }
      }
    }
  }
    list(Z, z)
}




# sets the entry in z corresponding to exp effect t(from-->to||seti) to TRUE (i.e. known to be zero)

add2zvec = function(z, from, to, seti, T, n,msg=NA, verbose = FALSE){
  id<-which( T[,2] == from & T[,3] == to & T[,4] == cset2idx(seti,n) ) 
  if ( verbose && !is.na(msg) ) {
    cat('%%',rownames(T)[id],'=0',' by ',msg,'\n', verbose)
  }
  z[id] = 1			
  z
}


# sets the two entries in Z corresponding to the product of exp effects 
# t(from1-->to1||seti1) t(from2-->to2||seti2) to TRUE (i.e. known to be zero)

add2Zmat = function(Z,from1,to1,seti1,from2,to2,seti2,T,n, msg=NA, verbose = FALSE){
  id1<-which( T[,2] == from1 & T[,3] == to1 & T[,4] == cset2idx(seti1,n) ) 
  id2<-which( T[,2] == from2 & T[,3] == to2 & T[,4] == cset2idx(seti2,n) ) 
  
  if (verbose && !is.na(msg) ) {
    cat('%%',rownames(T)[id1],'*',rownames(T)[id2],'=0',' by ',msg,'\n') 
  }
  
  Z[id1, id2] = TRUE
  Z[id2, id1] = TRUE
  Z
}


# converts a set of variables to the index of the set for look-ups in the parameter table T

cset2idx = function(seti, n){
  cset = rep(0,n)
  cset[seti] = TRUE
  set = bin.to.dec(cset) 
  set
}

