
# Uses the faithfulness rules described in the appendix to determine which experimental effects and which
# products of experimental effects are zero.


# INPUT
# Z: square matrix with n(n-1)2^(n-2) columns and rows 
#    Z[i,j] = 1 for those entries for which it is known that the product of the experimental effects 
#    represented by the row and column indices is zero, and which is zero otherwise
# z: ( n(n-1)2^(n-2))-ary vector which is 1 if the exp effect of the entry is known to be zero, and zero otherwise
# Call: covariance matrix over all variables in V for the exp-eriment
# exp: n-ary vector of {0,1,2} indicating the role of each variable in the experiment (hidden, obs, interv)
# nsamples: number of samples in exp
# maxCsets: maximum size of the conditioning set for minimal independence tests (the procedure 
# 	    	    will still check dependence using conditioning sets of size maxCsets+1 when possible)
# signif1: level of significance to establish independence: pval>signif1
# signif2: level of significance to establish dependence: pval<=signif2
# verbose: boolean controlling the commentary

# OUTPUT:
# returns a list containing Z and z after they have been updated using the implications from faithfulness 
# derived from this experiment


faithConst4Lininf = function(Z, z, Call, exp, nsamples, maxCsets, signif1, signif2,verbose=FALSE){
  n = length(exp)
  evec <- (exp==2)
  eveci <- which(evec)
  evecobs <- (exp==1)
  evecobsi <- which(evecobs)
  evecnothidden <- (exp>0)
  evecnothiddeni <- which(evecnothidden)  
  evechidden <- (exp == 0) 
  evechiddeni <- which(evechidden) 
  
  T = createParameterTable(n)	# lookup table to find index of a particular experimental effect
  
                                        # get all the conditioning sets that make a pair of variables (conditionally) independent
  sepsets = allminSepsets(Call, evecnothiddeni, nsamples, maxCsets, signif1)
    
                                        # consider independencies x_||_y | C with x,y in U first
  
  if(length(evecobsi)>=2){
    
    for(i in 1:(length(evecobsi)-1)){
      for(j in (i+1):length(evecobsi)){
        x = evecobsi[i]
        y = evecobsi[j]
        
                                        # consider all possible sepsets of x and y, if any	    
        
        if(length(sepsets[[x]][[y]])>0){
          
          for(ssi in 1:length(sepsets[[x]][[y]])){
            
            
            if(length(sepsets[[x]][[y]][[ssi]])==1 && (sepsets[[x]][[y]][[ssi]]) == 0){
              
                                        # x and y are unconditionally independent
              
              z = add2zvec(z,x,y,c(eveci, x),T,n, '1.1', verbose) 
              z = add2zvec(z,y,x,c(eveci, y),T,n, '1.2', verbose)
              
              for(u in c(evechiddeni, setdiff(evecobsi, c(x,y)))){
                Z = add2Zmat(Z,x,u,c(eveci, x),u,y,c(eveci,u),T,n, '1.5', verbose)
                Z = add2Zmat(Z,y,u,c(eveci, y),u,x,c(eveci,u),T,n, '1.6', verbose)
              }
              
              for(u in setdiff(1:n, c(x,y))){
                Z = add2Zmat(Z,u,y,c(eveci, u),u,x,c(eveci,u),T,n, '1.9', verbose)
              }
              
              
                                        # check for colliders w in U between x and y
              
              for(w in setdiff(evecobsi, c(x,y))){
                pval = isIndependent(Call, x, y, given = w, nsamples)
                if(pval<=signif2){
                  
					# x and y are unconditionally independent, but dependent conditional on some w in U
                  
                  z = add2zvec(z,w,y,c(eveci,w),T,n, '2.1', verbose)
                  z = add2zvec(z,w,x,c(eveci,w),T,n, '2.2', verbose)
                  
                  for(u in c(evechiddeni, setdiff(evecobsi, c(x,y,w)))){
                    Z = add2Zmat(Z,w,u,c(eveci, w),u,y,c(eveci, u),T,n, '2.4', verbose)
                    Z = add2Zmat(Z,w,u,c(eveci, w),u,x,c(eveci, u),T,n, '2.5', verbose)
                  }
                }
                
                
                if(pval>signif1){
                  
					#  x and y are unconditionally independent, and remain independent conditional on some w in U
                  
                  Z = add2Zmat(Z,x,w,c(eveci, x),y,w,c(eveci, y),T,n, '3.1', verbose)
                  
                }
              }
            }else{		# x and y are conditionally independent
              
              cset = (sepsets[[x]][[y]][[ssi]])
              
              z = add2zvec(z,x,y,union(c(eveci,x), cset),T,n, '1.1 with C', verbose)
              z = add2zvec(z,y,x,union(c(eveci,y), cset),T,n, '1.2 with C', verbose)
              
              for(u in c(evechiddeni, setdiff(evecobsi, c(x,y,cset)))){
                Z = add2Zmat(Z,x,u,union(c(eveci, x),cset),u,y,union(c(eveci,u),cset),T,n, '1.5 with C', verbose)
                Z = add2Zmat(Z,y,u,union(c(eveci, y),cset),u,x,union(c(eveci,u),cset),T,n, '1.6 with C', verbose)
              }
              
              for(u in setdiff(1:n, c(x,y,cset))){
                Z = add2Zmat(Z,u,y,union(union(eveci, u),cset),u,x,union(union(eveci,u),cset),T,n, '1.9 with C', verbose)
              }
              
              
                                        # check for colliders w in U\C between x and y
              
              for(w in setdiff(evecobsi, c(cset,x,y))){
                pval = isIndependent(Call, x, y, given = c(cset,w), nsamples)		
                if(pval<=signif2){
                  
                                        # x and y are independent conditional on cset, but dependent conditional on cset \cup w
                  
                  z = add2zvec(z,w,y,c(eveci,w),T,n, '2.1 with C', verbose)
                  z = add2zvec(z,w,x,c(eveci,w),T,n, '2.2 with C', verbose)
                  
                  for(c in setdiff(cset,eveci)){
                    z = add2zvec(z,w,c,c(eveci,w),T,n, '2.3', verbose)
                  }					
                  
                  for(u in c(evechiddeni, setdiff(evecobsi, c(cset,x,y,w)))){
                    Z = add2Zmat(Z,w,u,c(eveci, w),u,y,c(eveci, u),T,n, '2.4 with C', verbose)
                    Z = add2Zmat(Z,w,u,c(eveci, w),u,x,c(eveci, u),T,n, '2.5 with C', verbose)
                    
                    for(c in setdiff(cset,eveci)){
                      Z = add2Zmat(Z,w,u,c(eveci, w),u,c,c(eveci, u),T,n, '2.6', verbose)
                    }
                  }    		
                }
                
                
                if(pval>signif1){
                  
                                        #  x and y are conditionally independent, and remain independent conditional on cset cup w for some w in U
                  
                  Z = add2Zmat(Z,x,w,union(c(eveci, x),cset),y,w,union(c(eveci, y),cset),T,n, '3.1 with C', verbose)
                  
                }
              }
            }
          }
        }
      }
    }
  }
    
                                        # consider x_||_y | C when x in J and y in U
  
  if(length(eveci)>0 && length(evecobsi)>0){
    
    for(i in 1:length(eveci)){
      for(j in 1:length(evecobsi)){
        x = eveci[i]
        y = evecobsi[j]
        
                                        # consider all possible sepsets of x and y, if any	    
        
        if(length(sepsets[[x]][[y]])>0){
          
          for(ssi in 1:length(sepsets[[x]][[y]])){
            
            if( length(sepsets[[x]][[y]][[ssi]])==1 && (sepsets[[x]][[y]][[ssi]]) == 0){
              
                                        # x and y are unconditionally independent
              
              z = add2zvec(z,x,y,eveci,T,n, '1.1 with x in J', verbose)
              
              for(u in c(evechiddeni,setdiff(evecobsi, y))){
                Z = add2Zmat(Z,x,u,eveci,u,y,c(eveci,u),T,n, '1.5 with x in J', verbose)
              }
              
                                        # check for colliders w in U between x and y
              
              for(w in setdiff(evecobsi, y)){
                pval = isIndependent(Call, x, y, given = w, nsamples) 
                if(pval<=signif2){
                  
					# x and y are unconditionally independent, but dependent conditional on w
                  
                  z = add2zvec(z,w,y,c(eveci,w),T,n, '2.1 with x in J', verbose)
                  for(u in c(evechiddeni, setdiff(evecobsi, c(x,y,w)))){
                    Z = add2Zmat(Z,w,u,c(eveci, w),u,y,c(eveci, u),T,n, '2.4 with x in J', verbose)
                  }
                }
                
                
                if(pval>signif1){
                  
					#  x and y are unconditionally independent, and remain independent conditional on some w in U
                  
                  Z = add2Zmat(Z,x,w,eveci,y,w,c(eveci, y),T,n, '3.1 with x in J', verbose)
                  
                }
                
              }
            }else{			# x and y are independent conditional on cset
              
              cset = (sepsets[[x]][[y]][[ssi]])
              
              z = add2zvec(z,x,y,union(eveci,cset),T,n, '1.1 with C, and x in J', verbose)    
              
              for(u in c(evechiddeni, setdiff(evecobsi, c(y, cset)))){
                Z = add2Zmat(Z,x,u,union(eveci, cset),u,y,union(c(eveci,u),cset),T,n, '1.5 with C, and x in J', verbose)	
              }
              
              if(length(cset)==1){			#cset is a singleton conditioning set
                c = cset[1]
                if(exp[c]==1){				# c is observed
                  z = add2zvec(z,y,c,c(eveci,y),T,n, '1.3', verbose)
                  
                  for(u in c(evechiddeni,setdiff(evecobsi, c(y,c)))){
                    Z = add2Zmat(Z,y,u,c(eveci, y),u,c,c(eveci, u),T,n, '1.7', verbose)
                  }
                }
              }
              
              
                                        # check for colliders w in U\C between x and y
              
              for(w in setdiff(evecobsi, c(cset,y))){
                
                pval = isIndependent(Call, x, y, given=c(cset,w), nsamples)		
                if(pval<=signif2){
                  
                                        # x and y are independent conditional on cset, but dependent conditional on cset \cup w
                  
                  z = add2zvec(z,w,y,c(eveci,w),T,n, '2.1 with C, and x in J', verbose)
                  
                  for(c in setdiff(cset, eveci)){
                    z = add2zvec(z,w,c,c(eveci,w),T,n, '2.3 with x in J', verbose)
                  }
                  
                  for(u in c(evechiddeni, setdiff(evecobsi, c(cset,y,w)))){
                    Z = add2Zmat(Z,w,u,c(eveci, w),u,y,c(eveci, u),T,n, '2.4 with C, and x in J', verbose)
                    
                    for(c in setdiff(cset, eveci)){
                      Z = add2Zmat(Z,w,u,c(eveci, w),u,c,c(eveci, u),T,n, '2.6 with x in J', verbose)
                    }
                  }			
                }
                
                if(pval>signif1){
                  
                                        #  x and y are conditionally independent, and remain independent conditional cset cup w for some w in U
                  
                  Z = add2Zmat(Z,x,w,union(eveci, cset),y,w,union(c(eveci, y),cset),T,n, '3.1 with C, and x in J', verbose)
                  
                }
              }
            }
          }
        }
      }

      for(j in setdiff((1:length(eveci)),i)){
        
                                        # x and y are both intervened, and independent conditional on some w in U
                                        # this step is redundant if experimental effects are picked up elsewhere
        
        x = eveci[i]
        y = eveci[j]
        
        for(w in evecobsi){
          pval =   isIndependent(Call, x, y, given=w, nsamples)		
          if(pval>signif1){
            
            Z = add2Zmat(Z,x,w,eveci,y,w,eveci,T,n, '3.1 with x,y in J', verbose)
            
          }
        }
      }      
    }
  }
  list(Z, z)
}


# sets the entry in z corresponding to exp effect t(from-->to||seti) to TRUE (i.e. known to be zero)

add2zvec = function(z, from, to, seti, T, n,msg=NA, verbose = FALSE){
  id<-which( T[,2] == from & T[,3] == to & T[,4] == cset2idx(seti,n) ) 
  if ( verbose && !is.na(msg) ) {
    cat('%%',rownames(T)[id],'=0',' by ',msg,'\n')
  }
  z[id] = 1			
  z
}


# sets the two entries in Z corresponding to the product of exp effects 
# t(from1-->to1||seti1) t(from2-->to2||seti2) to TRUE (i.e. known to be zero)

add2Zmat = function(Z,from1,to1,seti1,from2,to2,seti2,T,n, msg=NA, verbose = FALSE){
  id1<-which( T[,2] == from1 & T[,3] == to1 & T[,4] == cset2idx(seti1,n) ) 
  id2<-which( T[,2] == from2 & T[,3] == to2 & T[,4] == cset2idx(seti2,n) ) 
  
  if (verbose && !is.na(msg) ) {
    cat('%%',rownames(T)[id1],'*',rownames(T)[id2],'=0',' by ',msg,'\n')
  }
  
  Z[id1, id2] = TRUE
  Z[id2, id1] = TRUE
  Z
}


# converts a set of variables to the index of the set for look-ups in the parameter table T

cset2idx = function(seti, n){
  cset = rep(0,n)
  cset[seti] = TRUE
  set = bin.to.dec(cset) 
  set
}
