#The first version of lin, in which fredericks stuff will be added

lininf<-function(DATA,nonzero_significance=0.01,zero_significance=0.1, true_values=NA,
            #  max_limit=Inf,
	    max_rounds=Inf,times=20,maxCset=Inf,verbose=1,faithful=TRUE,mode='lininf',plot=FALSE) { #max_limit nrow(DATA[[1]]$Cx)-2

  #DATA - is a list of data structures where each has the form
  #    list(Cx=,X=,N=,e=)
  # 	Cx is the covariance matrix, X are the samples, N is the number of samples (may be Inf),
  # 	is a vector defining the experiment (see data.createExperiments.R)
  #mode - describes whether we should operate like in 'aistats', 'pgm', or 'uai' algorithm
  #   - heh is a simulation of Hyttinen et al. "Causal Discovery for Linear Cyclic Models with Latent Variables" in PGM 2010
	  #-faithfulness may be utilized or not, for paper tests it is used
	  #-linear equations on direct effects
  #   - ehs is a simulation of Eberhardt et al. "Combining experiments..." in AISTATS 2009 ( published 2010)
	  #-faithfulness not utilized
	  #-linear equations on total effects
  #   - lininf is the linear inference algorithm from Hyttinen et al. "Causal Discovery...." in UAI 2012
  # nonzero_significance, zero_significance - input to algorithms in independence/, see those functions for details.
  # true_values - is a vector of true values for all experimental effects. Used for various printing options.
  # maxCset - the maximum conditioning set size used. See independence/ for details.
  # max_rounds - how many rounds to run the algorithms, limited for sample data and the simulated algorithms.
  # times - how many times to resample, a technique for dealing with sample data
  #
  # AISTATS <- NO FAITHFULNESS, ONLY EQUATIONS WITH TOTAL EFFECTS
  # PGM <- FAITHFULNESS may be TRUE or FALSE, ONLY EQUATIONS WITH DIRECT EFFECTS CONSIDERED


  #Checking that now 
  if ( mode == 'ehs' && faithful) {
    cat('THERE WERE NO FAITHFULNESS RULES IN AISTATS.\n')
    return(NA)
  }

  if ( mode == 'ehs' && max_rounds != 2 ) {
    #for aistats the linear inference procedure is run twice 
    # in the first round we have linear equations on the totals and solve for totals
    # in the second round we have linear equations on the directs with known totals
    cat('AISTATS SHOULD BE RUN FOR 2 ROUNDS.\n')
    return(NA)
  }

  if ( mode == 'heh' && max_rounds != 1 ) {
    cat('PGM SHOULD BE RUN FOR 1 ROUNDS.\n')
    return(NA)
  }

  #number of variables
  n<-length(DATA[[1]]$e)

  #resampling (bootstrapping) data, creating many data sets that are used in parallel 
  DATA<-resample(DATA, times)

  #number of experimental effects
  n_effects<-n*(n-1)*2^(n-2)

  #knowlege now contains also the resampled values
  knowledge<-list( T=createParameterTable(n), Z=array(FALSE,c(n_effects,n_effects)),E=array(NA,c(n_effects,times)))
  #T is the parameter table, see createParameterTable.
  #Z is a matrix, Z[i,j]=TRUE if we know that the product of i:th and j:th experimental effect is zero.
  #Thus Z includes the bilinear constraints. If some experimental effect is known to be zero, it is noted
  # in the parameter table.
  # E is a matrix where rows are individual estimates for the experimental effects
  #here estimates relate to the resampled estimates

  #putting the true valued to the parameter table for printing
  knowledge$T[,9]<-true_values

#############################################################################################
#######functions for knowledge########################################################
#############################################################################################

  #this function is used to add knowledge to the knowledge structure
  #knowledge is added as known experimental effects
  setknowledge<-function(knowledge,from,to,set,values,test) {
    #values is a list of estimates for the given experimental effect
    #if test=TRUE the exp. effect is tested for being zero or nonzero based on values

    #determine the index of the experimental effect
    i<-which( knowledge$T[,2] == from & knowledge$T[,3] == to & knowledge$T[,4] == set )

   #if already known do nothing!
   #test<-TRUE
    if ( knowledge$T[i,5] ) { #already known
        cat('-Not replacing knowledge:',rownames(knowledge$T)[i],'=',knowledge$T[i,6], 'with',
          mean(values),'\n')
    
        return(knowledge)
    }
    
    if ( test ) { #testing here with a randomization test whether 
      sig<-isZero(values)
      if ( sig > zero_significance ) {
        zero<-1
        nonzero<-0
      } else if ( sig < nonzero_significance ) {
        zero<-0
        nonzero<-1
      } else {
        zero<-0
        nonzero<-0
      }
    } else { #when inputting actual observations, the test have been done by code in independence
      zero<-0 #as no known values are ever set to anything
      nonzero<-1 #we get here only if the experimental effect has been nonzero!
      sig<-NA
    }


    true<-knowledge$T[i,9]

    cat('Adding knowledge:',rownames(knowledge$T)[i],'=',mean(values),
        'true=',true,'error:',true-mean(values),'zero=', zero, 'nonzero=', nonzero ,'by', sig ,'\n')

    knowledge$T[i,5]<-TRUE
    knowledge$T[i,6]<-mean(values)
    knowledge$T[i,7]<-zero
    knowledge$T[i,8]<-nonzero #not touching zeroes at all!

    knowledge$E[i,]<-values


    # here should iterate the consequences if a test was performed!
    # quite probably they will be iterated anyway in later rounds...
    #set this also by faithfulness
    # if a experimental effect is zero and faithfulness is used,
    # we know that the experimental effect with any superset of intervention set is also zero
    if ( zero && faithful && mode != 'ehs' && mode != 'heh' ) {

      for ( k in which( knowledge$T[,2] == from & knowledge$T[,3] == to ) ) { #go through parameters from->to
        if ( all( dec.to.bin(knowledge$T[k,4],n) >= dec.to.bin(set,n) ) && 
             ( knowledge$T[k,4] != set ) ) { #if the set is a superset 
          
          knowledge$T[k,5]<-TRUE
          knowledge$T[k,6]<-0
          knowledge$T[k,7]<-zero
          knowledge$T[k,8]<-nonzero

          knowledge$E[k,]<-rep(0,times)

          #set also pairs to zero NOT USEFULL, THEY WONT BE USED ANYWAY!
          #knowledge$Z[k,]<-TRUE
          #knowledge$Z[,k]<-TRUE                                     
          cat('Adding knowledge:',rownames(knowledge$T)[k],'=',0,
              'true=',true,'error:',true-0,'zero=', zero,  'nonzero=', nonzero ,'by', sig ,'\n')
        }
      }
    }
    knowledge
  }

###############################################################################################################
###############################################################################################################

  if ( faithful && mode != 'ehs' ) {
    cat('Using Independence code...\n')
    for ( i in 1:length(DATA) ) {
      e<-DATA[[i]]$e

      cat('Processing experiment: {',which(e==(1)),'},{',which(e==(0)),'},{',which(e==(-1)),'}\n')
      cat('CovarianceMatrix:\n')
      print(DATA[[i]]$Cx)

      J<-which(e==1); U<-which(e==0); L<-which(e==(-1));

############################################################################################
      nexpeffect<-nrow(knowledge$T)

      #Z[i,j]=TRUE if the product of the experimental effects is zero
      #z[i] =TRUE if i:th exp. effect is zero
    #starting with these 
      Zstart<- knowledge$Z 
      zstart<-knowledge$T[,7] == 1 


      if ( mode == 'lininf') { #for the linear inference algorithm

	Fresult<- faithConst4Lininf(Z=Zstart,z=zstart,
			    Call=DATA[[i]]$Cx,exp=DATA[[i]]$e+1, #indepence code uses a different experiment vector type
			    nsamples=DATA[[i]]$N,
		    maxCsets=maxCset,signif1=zero_significance,signif2=nonzero_significance)
	#this call returns two matrices 

	#postprocessing still the result
	Fresult2<-lininf.postprocess(Fresult[[2]], Fresult[[1]], knowledge$T, n)

      } else if ( mode == 'heh') { #for pgm algorithm simulation in the overlapping setting

	Fresult<- faithConst4Heh(Z=Zstart, z=zstart,
				      Call=DATA[[i]]$Cx,exp=DATA[[i]]$e+1,nsamples=DATA[[i]]$N,
				      maxCsets=maxCset,signif1=zero_significance,signif2=nonzero_significance)
	#no post processing, but changing the order 
	Fresult2<-list(Fresult[[2]], Fresult[[1]])
      }



      #####################JUST FOR PRINTING################################################################################
      cat('New knowledge from Independence code:\n')
      newz<-which( Fresult2[[1]]==1 & knowledge$T[,7]==0 )
      for ( j in index(1,length(newz)) ) {
	cat('New zero:',rownames(knowledge$T)[newz[j]],'=0, true:',true_values[newz[j]],'\n')
      }

      newZ<-which( Fresult2[[2]] & !knowledge$Z, arr.ind=TRUE)
      for ( j in index(1,nrow(newZ)) ) {
	cat('New bilinear zero:',rownames(knowledge$T)[newZ[j,1]],'*',rownames(knowledge$T)[newZ[j,2]],
		  '=0, true:',true_values[newZ[j,1]]*true_values[newZ[j,2]],'\n')
      }

    ######################################################################################################################

      #simply adding it in
      #for now add it here that any product with a zero exp effect is zero.

      #setting to zeroes
      knowledge$T[Fresult2[[1]] == 1,7]<- 1   
      #setting the estimates to zero
      knowledge$T[Fresult2[[1]] == 1,6]<-0
      #setting the bootstrapped estimates to zero
      knowledge$E[Fresult2[[1]] == 1,]<-0
      #setting the values as known
      knowledge$T[Fresult2[[1]] == 1,5]<-1

      #adding bilinear zeroes
      knowledge$Z<-knowledge$Z | Fresult2[[2]]


      #cat('Check results...\n')
      #browser()
      #It is not useful to set these to zero, they wont be used anyway!
      #knowledge$Z[knowledge$T[,7]==1,]<-TRUE
      #knowledge$Z[,knowledge$T[,7]==1]<-TRUE
    }
  }
###########################################################################################

  cat('Adding in the known experimental effects...\n\n')

  for ( i in 1:length(DATA) ) {
    e<-DATA[[i]]$e

    cat('Processing experiment: {',which(e==(1)),'},{',which(e==(0)),'},{',which(e==(-1)),'}\n')
    cat('CovarianceMatrix:\n')
    print(DATA[[i]]$Cx)

    J<-which(e==1); U<-which(e==0); L<-which(e==(-1));

    for ( j in J ) {
      for ( u in U ) {
         if ( faithful ) { #when using faithfulness tests whether a element is zero or non-zero_significance
			   #have already been done by the independence/ code
            knowledge<-setknowledge(knowledge,j,u,bin.to.dec( 1*(e==1)),DATA[[i]]$R[u,j,],test=FALSE) 
         } else {
            knowledge<-setknowledge(knowledge,j,u,bin.to.dec( 1*(e==1)),DATA[[i]]$R[u,j,],test=TRUE) 
         }
      }
    }
  }
###############################################################################################
###############################################################################################
  if( n < 5 ) {
    cat('Knowledge after observations:\n')
    print(knowledge$T)
  }

  cat('Max error:',max(abs(knowledge$T[,6]-knowledge$T[,9]),na.rm=TRUE),'\n')


####################################################################################################
################################LINEAR INFERENCE####################################################
####################################################################################################

  cat('Next iterating knowledge...\n')

  if (verbose >= 2 ) browser()

  round<-1
  while ( round <= max_rounds ) {
    cat('Round:',round,'/',max_rounds,'\n')

    #creating equations that are linear on the unknowns
    equations<-createLinearEquations( n, knowledge, limit=Inf, verbose=verbose, mode=mode, round=round)

    #indexes for known and unknown elements
    known<-which(knowledge$T[,5]==1) 
    unknown<-which(knowledge$T[,5]==0) # & !unconst
    n_unknown<-length(unknown)


    if ( nrow(equations$A[[1]]) == 0  ) {
      cat('No more usable equations :(\n')
      if( verbose >= 2 ) browser()
      break
    }

    for ( t in 1:times ) {
      #keep only the colums corresponding to unknows, deleting others, such that the system is smaller
      equations$A[[t]]<-equations$A[[t]][,unknown,drop=FALSE] 
      if (nrow(equations$A[[t]]) > 0 ) {
        cat('Equations are still satisfied by true model:',sum((equations$A[[t]]%*%true_values[unknown] - equations$b[[t]])^2),'\n')
      }
    }

    x<-array(NA,c(n_unknown,times) ) #esimated values for the unknowns

    x_det<-array(NA,c(n_unknown,n_unknown,times) )
    x_det2<-array(NA,c(n_unknown,times) )

    #solving 
    for ( t in 1:times ) {
      Apinv<-NA
      try( {Apinv<-mpinv(equations$A[[t]])}, silent = !verbose )  

      if ( any(is.na(Apinv) ) ) next

      
      #solving the system for t:th resampling round
      x[,t]<-Apinv%*%equations$b[[t]]


      #now we still have to find which elements of x[,t] are actually uniquely determined by the equation system
      # and which are not?

      #rows of r are now the residuals when projecting to the row space of A

      if ( is.infinite( DATA[[1]]$N ) ) { 
	#if N is infinite there is trouble of sample variance
	#this calculates a projection of the null space vectors???????????????

	#this project vector e_i into the space of null-space vectors
	#Here we should figure out which elements of the linear equation system are determined and which are not.

	#the i:th element is uniquely determined if the null space vectors have all zero on the ith coordinates

	#this is done followingly: 
	# first projecting e_i to row space of A, column space of t(A)
	# e_i= t(A)%*%gamma
	# gamma= t(Apinv)%*%e_i
	# e_i_projected = t(A)%*%t(Apinv)%*%e_i
	# Now the e_i projected into the null space is e_i-e_i_projected
	# t(e_i-e_i_projected) = t(e_i) -t(e_i)%*%Apinv%*%A = t(e_i) %*% ( eye - Apinv%*%A )
	# the thing below does this for all vectors
	# thus x_det[i,,t] is e_i projected into the null space in the t:th resampling round

        x_det[,,t]<-diag(nrow(Apinv))-Apinv%*%equations$A[[t]]

	#very nice, this can be calculated without explicitly but unfortunately this does not work with sample data
      } else {
        #this is done with sample data
	# let  there be prior on x N(x;0,sigma_prior*eye)
	# likelihood of the linear data is N(A*x-b;0,sigma_data*eye)
	# multiplying we get posterior for x, N(x;?,inv( 1/(2*sigma_data)*t(A)%*%A +1/(2*sigma_prior)*eye ) )
	# thus x_det2[i,t] is a posterior variance of i:th element
	# if it is below a threshold the element is uniquely determined
	# this bayesian way seems to work best here

	  try( {x_det2[,t]<-abs( diag( mpinv( 1/(2*1e-1)*t( equations$A[[t]] ) %*% equations$A[[t]]
				  + 1/(2*1e2)*diag(ncol(equations$A[[t]]) ) )))}, silent = !verbose )  
	  if (any( is.na(x_det2[,t]) ) ) next
      }
    }


    #the following determines the newly known parameters in two settings
    # first with infinite sample data, second with sampledata
    if ( is.infinite ( DATA[[1]]$N )  ) {
      coord_is_zero_pval<-apply(x_det,c(1,2),isZero) #determine which coordinates are effectively zero

      newknown<- !apply( coord_is_zero_pval < nonzero_significance,2,any)
    } else {
      #which elements are determined with sample data?

      threshold<-1

      #errors are calculated for printing
      errors<-  abs( true_values[unknown] - rowMeans(x,na.rm=TRUE) )


      #calculating the mean variance over the resampled data sets
      vars<-rowMeans(x_det2,na.rm=TRUE)

      #if the posterior variance is under a thereshold, the exp. effect is considered determined
      id <- vars < threshold 


      if ( plot ) { #this plotting still shows the situation in when using sample data
	  logvar_id<-log(vars[id],base=10)
	  error_id<-errors[id]
      
	  logvar_nid<-log(vars[!id],base=10) 
	  error_nid<-errors[!id]

          #plot the errors and which are identified and which are not
          plot( c(logvar_id,logvar_nid),c(error_id,error_nid),type='p',pch='x',col='black',main='Errors of the parameters'
                ,xlab='log variances', ylab='abs. errors')
          points(logvar_id,error_id,pch='o',col='green')
          points(logvar_nid,error_nid,pch='o',col='red')
      
          lines( c(1,1)*log(threshold,base=10), 10000*c(-1,1), col='blue')
      
          legend('topleft', col=c('blue','green','red'),legend=c('theshold line','identified','underdetermined'),pch=c('-','o','o'))
      }

      newknown <- id
    }

    if ( all(!newknown) ) {
      cat('Nothing more found :(\n')
      if( verbose >= 2 ) browser()
      break
    }

    #Addinf the new known elements to knowledge structure
    for ( i in which(newknown) ) {
      cat('New knowledge on parameter',rownames(knowledge$T)[unknown[i]],'!\n')
      #browser()
      #here should test for zeroness and make the conclusions based on faithfulness
      knowledge<-setknowledge(knowledge,knowledge$T[unknown[i],2],knowledge$T[unknown[i],3],
                              knowledge$T[unknown[i],4],x[i,!is.na(x[i,])],test=TRUE)

    }
    #Printing out 
    cat('Max error:',max(abs(knowledge$T[unknown[newknown],6]-true_values[unknown[newknown]])),'\n')
    if ( n < 5 ) {
      cat('Knowledge after ',round,'round:\n')
      print(knowledge$T)
    }

    if ( round == max_rounds) {
        cat('Maximum number of rounds:',round,'achieved.','\n')
        if( verbose >= 2 ) browser()
        break
    }


    if (verbose >= 2 ) browser()
    round<-round+1
    cat('\n')
  }
  if ( n < 5 ) {
    cat('Finally:\n')
    print(knowledge$T)
  }
  #browser()
  #finally returning all knowledge
  knowledge
}